/************************************************************************/
/* conf_read.y: This file contains grammars and parsing routines for    */
/*          formulating search filters.					*/	
/*                                                                      */
/* Modification Log:                                                    */
/*                                                                      */
/* 91/05/21     Sze-Ying Wuu: additions/changes made for xdi, v 1.0     */
/* 91/02/22     /f/osi/others/quipu/uips/pod/RCS/conf_read.y,v 7.2      */
/*              Isode Interim (6.8)                                     */
/************************************************************************/

/* $Header: conf_read.y,v 1.1 91/06/24 14:23:01 sywuu Exp $ */

/*
 *    Original source: ISODE Interim Release 6.8
 *    Changes/additions by: Sze-Ying Wuu, Bellcore, <sywuu@thumper.bellcore.com>
 *
 *               Changes and Additions Copyright (c) 1991 Bellcore
 *                            All Rights Reserved
 *
 *    Permission is granted to copy, modify and redistribute this program
 *    in source or binary form provided that it is not sold for
 *    profit, and all copyright notices, permission and nonwarranty notices 
 *    remain intact.
 *
 *    Bellcore makes no warranty and accepts no liability for this program.
 * 
 *
 *                             ISODE NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

%{
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "quipu/util.h"
#include "quipu/entry.h"
#include "filt.h"

int make_filt_type();
int make_as_type();
filt_struct *make_item_filter();
filt_struct *link_filters();
filt_struct *make_parent_filter();
attr_type *make_attr_type();
Attr_Sequence make_as_item();
Attr_Sequence merge_as_seq();

extern FILE *config_file;
extern unsigned int curr_filt;
extern char *file_names[], *entrytype[];
%}

/* token may be one of these types */
%union {
  filt_struct *filt;
  attr_type *atype;
  Attr_Sequence as;
  char strval[255];
  int symbol;
}

%token NUMBER NAME DEFAULT STRING OID AND OR NOT APPROX EQUAL GREATEREQUAK LESSEQUAL ITEM 
%type <filt> filter filter_list assertion filter_item
%type <as> attribute attribute_seq
%type <symbol> filt_type match
%token <symbol> NOT AND OR APPROX EQUAL SUBSTRING GREATEREQUAL LESSEQUAL
%token <symbol> '"' ':' '(' ')' '{' '}'
%token <strval> NAME DEFAULT STRING OID
%type  <atype> name default

/* parsing grammar */
/* attr_spec: grammar specification for attribute sequence */
/* filt_spec: grammar specification for filter */
%%
 top          : spec_list
              | /* empty */			{return(1);}
              ;

 spec_list    : spec_item
              | spec_list spec_item
            
 spec_item    : attr_spec
              | filt_spec

 attr_spec    : name '{' attribute_seq '}'	{make_as_type($1, $3);}
              ;

 filt_spec    :	name filter			{make_filt_type($1, $2);}
  	      ;	

 name         :	NAME ':' STRING			{$$ = make_attr_type($1, $3);}
	      ;

 default      :	DEFAULT ':' STRING		{$$ = make_attr_type($1, $3);}
              |					{(void) strcpy($$, "\0");}
	      ;

 attribute_seq: attribute attribute_seq		{$$ = merge_as_seq($1, $2);}
              | attribute 			{$$ = $1;}
              |					{$$ = NULLATTR;}
              ;

 attribute    : OID				{$$ = make_as_item($1);}
              ;

 assertion    : '(' filt_type filter filter filter_list ')' 	{$$ = make_parent_filter($2, $3, $4, $5);}
	      | '(' NOT filter ')' 				{$$ = make_parent_filter($2, $3, (filt_struct *) 0,(filt_struct *) 0);}
	      | filter_item					{$$ = $1;}
	      ;

 filter_list  : filter filter_list 			{$$ = link_filters($1, $2);}
	      | filter					{$$ = $1;}
  	      |						{$$ = (filt_struct *) 0;}
	      ;

 filter       : filter_item			{$$ = $1;}
	      | assertion			{$$ = $1;}
	      ;

 filter_item  : '(' OID match STRING ')'	{$$ = make_item_filter($2, $3, $4);}
	      ;

 match        : APPROX				{$$ = $1;}
	      | EQUAL				{$$ = $1;}
	      | GREATEREQUAL                    {$$ = $1;}
	      | LESSEQUAL                       {$$ = $1;}
	      | SUBSTRING                       {$$ = $1;}
  	      ;

 filt_type    : AND				{$$ = $1;}
	      | OR				{$$ = $1;}
	      ;
%%

yylex()
{
  int c, count = 0;
  char lexeme[255];
  int colon = FALSE;
  
  while(isspace(c = getc(config_file)))
    if (c == EOF) return(0);
  
  lexeme[count++] = c;
	
  switch(c) {
  case '#':
    while (getc(config_file) != '\n');
    return(yylex());
  case '"':
    count = 0;
    while ((c = getc(config_file)) != '"')
      lexeme[count++] = c;
    lexeme[count] = '\0';
    (void) strcpy(yylval.strval, lexeme);
    return STRING;
  case '(':
    return (int) c;
  case ')':
    return (int) c;
  case '{':
    return (int) c;
  case '}':
    return (int) c;
  case ':':
    colon = TRUE;
    return (int) c;
  case '&':
    yylval.symbol = AND;
    return AND;
  case '|':
    yylval.symbol = OR;
    return OR;
  case '!':
    yylval.symbol = NOT;
    return NOT;
  case '*':
    lexeme[count] = '\0';
    (void) strcpy(yylval.strval, lexeme);
    return STRING;
  case '~':
    if((lexeme[count] = getc(config_file)) == '=') {
      yylval.symbol = APPROX;
      return APPROX;
    }		  	
    break;
  case '%':
    if((lexeme[count] = getc(config_file)) == '=') {
      yylval.symbol = SUBSTRING;
      return SUBSTRING;
    }
    break;
  case '>':
    if((lexeme[count] = getc(config_file)) == '=') {
      yylval.symbol = GREATEREQUAL;
      return GREATEREQUAL;
    }
    break;
  case '<':
    if((lexeme[count] = getc(config_file)) == '=') {
      yylval.symbol = LESSEQUAL;
      return LESSEQUAL;
    }
    break;
  case '=':
    yylval.symbol = EQUAL;
    return EQUAL;
  }
  
  while(!isspace(c = getc(config_file)) && c != '\0' && !issymbol(c))
    if (c != EOF)
      lexeme[count++] = c;
    else 
      return(0);
  
  (void) fseek(config_file,(long) -1, 1);

  lexeme[count] = '\0';
  switch(*lexeme) {
  case 'd':
  case 'D':
    if(!strcmp(lexeme, "default") || !strcmp(lexeme, "DEFAULT"))
      return DEFAULT;
    else {
      (void) strcpy(yylval.strval, lexeme);
      return STRING;
    }
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    count = 0;
    while (isdigit(lexeme[count]) || lexeme[count] == '.') count++;
    if (lexeme[count] == '\0') {
      (void) strcpy(yylval.strval, lexeme);
      return OID;
    } else {
      (void) strcpy(yylval.strval, lexeme);
      return STRING;
    }
  default:
    (void) strcpy(yylval.strval, lexeme);
    if (colon) {
       colon = FALSE;
       return STRING;
    }
    else 
       return NAME;
  }
}

/* ARGSUSED */
yyerror(err)
     char *err;
{
  (void) fprintf(stderr, 
		 "Parse error %s in '%s'. Exiting.\n", 
		 err, (char *) file_names[curr_filt]);
  if (entrytype[curr_filt]) {
    free(entrytype[curr_filt]);
    entrytype[curr_filt] = (char *) 0;
  }
  
  exit(1);
}

int issymbol(c)
char c;
{
  switch(c) {
  case '#':
  case '"':
  case '(':
  case ')':
  case '{':
  case '}':
  case ':':
  case '&':
  case '|':
  case '!':
  case '*':
  case '~':
  case '=':
  case '%':
  case '>':
  case '<':
    return 1;
  }
  return 0;
}
